#include <windows.h>
#include "AdrTime.h"

#include <stdlib.h>
#include <math.h>

extern "C" {
#include "mpg123.h"
}

static inline void SwapByte(BYTE& b1, BYTE& b2)
{
	BYTE temp = b1;
	b1 = b2;
	b2 = temp;
}

CAdrTime::CAdrTime()
{
	m_pData = NULL;
	m_hFile = NULL;
}

CAdrTime::~CAdrTime()
{
	Release();
}

BOOL CAdrTime::Init(const char* pFileName, int pos)
{
	Release();
	HANDLE hFile = ::CreateFile(pFileName, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE,
		NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFile == INVALID_HANDLE_VALUE)
		return FALSE;

	m_hFile = hFile;
	frame fr;
	::SetFilePointer(m_hFile, pos, NULL, FILE_BEGIN);
	if (GetFrame(fr) == FALSE)
	{
		::CloseHandle(m_hFile);
		m_hFile = NULL;
		return FALSE;
	}
	::SetFilePointer(m_hFile, -fr.framesize - 4, NULL, FILE_CURRENT);

	m_nMaxData = ::GetFileSize(m_hFile, NULL) / fr.framesize;
	m_nMaxData += m_nMaxData / 20;//5%]Ɋm
	m_pData = (AdrTimeData*)::malloc(m_nMaxData * sizeof(AdrTimeData));
	m_nNumData = 0;
	AddData(GetFilePtr(), 0.0);
	return TRUE;
}

void CAdrTime::Release()
{
	if (m_pData)
	{
		::free(m_pData);
		m_pData = NULL;
	}
	if (m_hFile)
	{
		::CloseHandle(m_hFile);
		m_hFile = NULL;
	}
}

DWORD CAdrTime::ToAddress(int nTime)
{
	if (m_hFile == NULL || nTime < 0)
		return -1;
	while ((m_pData + m_nNumData - 1)->nTime < nTime)
	{
		if (ProcessOneFrame() == FALSE)
			break;
	}
	if ((m_pData + m_nNumData - 1)->nTime <= nTime)
		return (m_pData + m_nNumData - 1)->dwAdr;

	AdrTimeData dummy[2];
	dummy[0].nTime = dummy[1].nTime = nTime;
	AdrTimeData* pData = (AdrTimeData*)::bsearch(
		dummy, m_pData, m_nNumData - 1, sizeof(AdrTimeData),
		(int (*)(const void*, const void*))AdrTimeData::CompTime);
	return pData->dwAdr;
}

int CAdrTime::ToTime(DWORD dwAdr)
{
	if (m_hFile == NULL || dwAdr < m_pData->dwAdr)
		return -1;
	while ((m_pData + m_nNumData - 1)->dwAdr < dwAdr)
	{
		if (ProcessOneFrame() == FALSE)
			return -1;
	}
	
	AdrTimeData* pData = (AdrTimeData*)::bsearch(
		&AdrTimeData(dwAdr, 0), m_pData, m_nNumData, sizeof(AdrTimeData),
		(int (*)(const void*, const void*))AdrTimeData::CompAdr);
	if (pData == NULL)
		return -1;
	return pData->nTime;
}

double CAdrTime::GetFrameTime(const frame& fr)
{
	static const double table[4][3] ={
		{384000.0, 1152000.0, 1152000.0},/*mpeg1  */
		{384000.0, 1152000.0,  576000.0},/*mpeg2  */
		{384000.0, 1152000.0,  576000.0} /*mpeg2.5*/
		/* layer1,    layer2,    layer3*/
	};
	return table[fr.mpeg25 * 2 + (fr.lsf >> fr.mpeg25)][fr.lay - 1] / (double)freqs[fr.sampling_frequency];
}

BOOL CAdrTime::GetFrame(frame& fr)
{
	int i;
	for (i = 0; i < 65536; i++)
	{
		DWORD dw = GetFilePtr();
		int r = ReadFrame(fr);
		if (i == 0 && r == 0)
			break;
		else if (r == 0)
		{
			DWORD dwCur = GetFilePtr();
			int j;
			for (j = 0; j < 4; j++)
			{
				frame f;
				if (ReadFrame(f) == -2)
					break;
			}
			if (j == 4)
			{
				::SetFilePointer(m_hFile, dwCur, NULL, FILE_BEGIN);
				break;
			}
			else
				::SetFilePointer(m_hFile, dw + 1, NULL, FILE_BEGIN);
		}
		else if (r == -1)
			return FALSE;
	}
	if (i == 65536)
		return FALSE;
	return TRUE;
}

// 0 
//-1 EOF
//-2 s
int CAdrTime::ReadFrame(frame& fr)
{
	union
	{
		DWORD header;
		BYTE byte[4];
	};
	DWORD nbr;
	::ReadFile(m_hFile, &header, 4, &nbr, NULL);
	if (nbr != 4)
	{
		::SetFilePointer(m_hFile, -(long)nbr, NULL, FILE_CURRENT);
		return -1;
	}
	SwapByte(byte[0], byte[3]);
	SwapByte(byte[1], byte[2]);
	if (CheckHeader(header) == FALSE || ::decode_header(&fr, header) == 0)
	{
		int i;
		for (i = 2; i >= 0 && byte[i] != 0xFF; i--);
		::SetFilePointer(m_hFile, - i - 1, NULL, FILE_CURRENT);
		return -2;
	}
	::SetFilePointer(m_hFile, fr.framesize, NULL, FILE_CURRENT);
	return 0;
}

BOOL CAdrTime::CheckHeader(DWORD header)
{
    if ((header & 0xffe00000) != 0xffe00000 ||
		((header >> 17) & 0x03) == 0x00 ||
		((header >> 12) & 0x0F) == 0x0F ||
		((header >> 10) & 0x03) == 0x03 )
		return FALSE;
	return TRUE;
}

DWORD CAdrTime::GetFilePtr()
{
	return ::SetFilePointer(m_hFile, 0, NULL, FILE_CURRENT);
}

BOOL CAdrTime::AddData(DWORD dw, double d)
{
	if (m_nNumData == m_nMaxData)
	{
		m_nMaxData *= 2;
		void* p = ::realloc(m_pData, m_nMaxData * sizeof(AdrTimeData));
		if (p == NULL)
			return FALSE;
		m_pData = (AdrTimeData*)p;
	}
	(m_pData + m_nNumData)->dwAdr = dw;
	(m_pData + m_nNumData)->nTime = (int)d;
	m_nNumData++;
	m_dTime = d;
	return TRUE;
}

BOOL CAdrTime::ProcessOneFrame()
{
	frame fr;
	if (GetFrame(fr) == FALSE)
		return FALSE;
	if (AddData(GetFilePtr(), m_dTime + GetFrameTime(fr)) == FALSE)
		return FALSE;
	return TRUE;
}

int CAdrTime::AdrTimeData::CompAdr(const AdrTimeData* p1, const AdrTimeData* p2)
{
	return p1->dwAdr - p2->dwAdr;
}

int CAdrTime::AdrTimeData::CompTime(const AdrTimeData* p1, const AdrTimeData* p2)
{
	if (p2->nTime == (p2 + 1)->nTime)
	{
		if (p1->nTime <= p2->nTime && p2->nTime < (p1 + 1)->nTime)
			return 0;
	}
	else if (p2->nTime <= p1->nTime && p1->nTime < (p2 + 1)->nTime)
		return 0;
	return p1->nTime - p2->nTime;
}
